/* ==================================================================
   File: OpenLoop.h

   This file serves as an example of how to use the command
   functions in "Commands.h".  It provides functions to save the
   aircraft data to an ASCII CSV file, plus open-loop control
   routines, of the kind that would commonly be used for flight
   dynamics studies, that send commands to the simulation to
   fly the simulated aircraft.

            Copyright (C) 2014 Richard C. Wagner
   This file is offered under the GNU General Public License.
   Please see the license statement at the end of the file.
   =============================================================== */

char OutFileName[] = "FlightData.csv";
FILE *OutFileHandle;


void SaveTitles()
{
   fprintf(OutFileHandle,"Time, u, v, w, p, q, r, Phi, Theta, Psi, Alpha,");
   fprintf(OutFileHandle,"Beta, Gamma, X, Y, Alt., delE, delA, delR, Throttle\n");
   return;
}

void SaveRateValues()
{
    fprintf(OutFileHandle,"%f, %f, %f, %f, %f, %f, %f,",
            *ar_simtime, *ar_u, *ar_v, *ar_w, RadiansToDegrees(*ar_p),
            RadiansToDegrees(*ar_q), RadiansToDegrees(*ar_r));
    return;
}

void SaveStateValues()
{
    fprintf(OutFileHandle,"%f, %f, %f, %f, %f, %f,",
            RadiansToDegrees(*ar_phi), RadiansToDegrees(*ar_theta), RadiansToDegrees(*ar_psi),
            RadiansToDegrees(*ar_alpha), RadiansToDegrees(*ar_beta), RadiansToDegrees(*ar_gamma));
    return;
}

void SavePositionValues()
{
    fprintf(OutFileHandle,"%f, %f, %f,", *ar_x, *ar_y, *ar_altitude);
    return;
}

void SaveDeflectionValues()
{
    fprintf(OutFileHandle,"%f, %f, %f, %f",
            RadiansToDegrees(*ar_dele), RadiansToDegrees(*ar_dela), RadiansToDegrees(*ar_delr),
            (*ar_throttle * 100));
    return;
}

// Writes the values returned in the advance response message to the data file.
void SaveValues()
{
    SaveRateValues();
    SaveStateValues();
    SavePositionValues();
    SaveDeflectionValues();
    fprintf(OutFileHandle,"\n");
    return;
}

// Creates a CSV data file and saves the column labels.  Returns true on success.
bool CreateTest(HWND hWND)
{
    bool result;

    OutFileHandle = fopen(OutFileName, "w+");
    if (OutFileHandle != 0)
    {   SaveTitles();
        result = true;
    }
    else
    {
        MessageBox(hWND, "Couldn't open the output file.", "Error", MB_OK);
        result = false;
    }
    return result;
}

// Closes the data file.
void CloseTest()
{
    fclose(OutFileHandle);
    return;
}


// ================  Open-Loop Control Routines  ====================

float DeltaT = 0.1;  // We'll exert control and save data every 0.1 sec.

// Advances the simulation "Time" seconds with the current control inputs.
// Saves data to the CSV file every DeltaT seconds.
void Fly(float Time)
{
    int i, Steps = (int) ((Time / DeltaT) + 0.5);
    double Rem = Time - (DeltaT * Steps);

    for (i = 0; i < Steps; i++)
    {
        AdvanceSimulation(DeltaT);
        SaveValues();
    }
    if (Rem > (DeltaT / 100))
    {
        AdvanceSimulation(Rem);
        SaveValues();
    }
    return;
}

// Flies the aircraft straight and level for 120 seconds.
void StraightAndLevel(HWND hWND)
{
    bool result;

    result = CreateTest(hWND);
    if (result == true)
    {
        Fly(120.0);
        CloseTest();
    }
    return;
}

// Executes a "stick rap" at 3 seconds.  Flies for a total of 120 seconds.
void StickRap(HWND hWND)
{
    bool result;

    result = CreateTest(hWND);
    if (result == true)
    {
        Fly(3.0);
        *ac_dele = *ac_dele + DegreesToRadians(8.0);
        Fly(1.0);
        *ac_dele = *ac_dele - DegreesToRadians(8.0);
        Fly(116.0);
        CloseTest();
    }
    return;
}

// Executes an elevator doublet at 3 seconds.  Flies for a total of 120 seconds.
void ElevatorDoublet(HWND hWND)
{
    bool result;

    result = CreateTest(hWND);
    if (result == true)
    {
        Fly(3.0);
        *ac_dele = *ac_dele - DegreesToRadians(4.0);
        Fly(1.0);
        *ac_dele = *ac_dele + DegreesToRadians(8.0);
        Fly(1.0);
        *ac_dele = *ac_dele - DegreesToRadians(4.0);
        Fly(115.0);
        CloseTest();
    }
    return;
}

// Executes an aileron doublet at 3 seconds.  Flies for a total of 120 seconds.
void AileronDoublet(HWND hWND)
{
    bool result;

    result = CreateTest(hWND);
    if (result == true)
    {
        Fly(3.0);
        *ac_dela = *ac_dela - DegreesToRadians(4.0);
        Fly(1.0);
        *ac_dela = *ac_dela + DegreesToRadians(8.0);
        Fly(1.0);
        *ac_dela = *ac_dela - DegreesToRadians(4.0);
        Fly(115.0);
        CloseTest();
    }
    return;
}

// Executes a rudder doublet at 3 seconds.  Flies for a total of 120 seconds.
void RudderDoublet(HWND hWND)
{
    bool result;

    result = CreateTest(hWND);
    if (result == true)
    {
        Fly(3.0);
        *ac_delr = *ac_delr - DegreesToRadians(4.0);
        Fly(1.0);
        *ac_delr = *ac_delr + DegreesToRadians(8.0);
        Fly(1.0);
        *ac_delr = *ac_delr - DegreesToRadians(4.0);
        Fly(115.0);
        CloseTest();
    }
    return;
}


/* ==================================================================
   Here's a set of routines that are semi-closed-loop.  They exert
   bang-bang control over the simulated aircraft.
   =============================================================== */

// A place to save the aircraft trim values
double dela_trim, dele_trim, delr_trim, throttle_trim;

void SaveTrimValues()
{
    dela_trim = *tr_dela;
    dele_trim = *tr_dele;
    delr_trim = *tr_delr;
    throttle_trim = *tr_throttle;
    return;
}

void RestoreTrimValues()
{
    *ac_dela = dela_trim;
    *ac_dele = dele_trim;
    *ac_delr = delr_trim;
    *ac_throttle = throttle_trim;
    return;
}

// Executes a loop at 3 seconds.  Flies until the loop is complete.
void StraightLoop(HWND hWND)
{
    bool result;

    result = CreateTest(hWND);
    if (result == true)
    {
        SaveTrimValues();
        Fly(3.0);
        *ac_dele = DegreesToRadians(-15.0);
        *ac_throttle = 1.0;
        Fly(DeltaT);
        do
            Fly(DeltaT);
        while ((*ar_gamma < DegreesToRadians(360.0)) && (*ar_gamma >= DegreesToRadians(0))
                && (*ar_simtime < 60.0));
        RestoreTrimValues();
        Fly(10.0);
        CloseTest();
    }
    return;
}

// Executes an aileron roll at 3 seconds.  Flies until the roll is complete.
void AileronRoll(HWND hWND)
{
    bool result;

    result = CreateTest(hWND);
    if (result == true)
    {
        SaveTrimValues();
        Fly(3.0);
        *ac_dela = DegreesToRadians(15.0);
        Fly(DeltaT);
        do
            Fly(DeltaT);
        while (*ar_phi <= DegreesToRadians(360.0));
        RestoreTrimValues();
        Fly(10.0);
        CloseTest();
    }
    return;
}


/* ====================  License Statement  =========================
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   =============================================================== */
