/* ===============================================================
   File: main.cpp

   This file acts as a load file and also provides the GUI for
   the flight simulation client.  The GUI is generated using
   a dialog box with no parent window.  The best reference for
   the code included here is probably Petzold's "Programming
   Windows", along with the Microsoft Developer's Network, at
   http://msdn.microsoft.com/en-us/windows/desktop
   Enter any function call you see here in the search field at
   the upper-right of the web page.

            Copyright (C) 2014 Richard C. Wagner
   This file is offered under the GNU General Public License.
   Please see the license statement at the end of the file.
   =============================================================== */

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdlib.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdbool.h>
#include <math.h>

#include "resource.h"
#include "ProgrammerAids.h"
#include "DataValues.h"
#include "NetworkSupport.h"
#include "Commands.h"
#include "OpenLoop.h"


// ====================  Global Variables  ==========================

HINSTANCE hInst;


// ====================  Utility Functions  =========================

void EnableControl(int ID, HWND hDLG)
{
    HWND hWND;

    hWND = GetDlgItem(hDLG, ID);
    EnableWindow(hWND, true);
    return;
}

void DisableControl(int ID, HWND hDLG)
{
    HWND hWND;

    hWND = GetDlgItem(hDLG, ID);
    EnableWindow(hWND, false);
    return;
}

void EnableTrimButtons(HWND hDLG)
{
    EnableControl(IDPB_TRIM_ALPHA, hDLG);
    EnableControl(IDPB_TRIM_SPEED, hDLG);
    EnableControl(IDPB_TRIM_DELE, hDLG);
    return;
}

void EnableFlyButtons(HWND hDLG)
{
    int i;

    for (i = IDPB_STRAIGHT_AND_LEVEL; i <= IDPB_AILERON_ROLL; i++ )
        EnableControl(i, hDLG);
    return;
}

void DisableFlyButtons(HWND hDLG)
{
    int i;

    for (i = IDPB_STRAIGHT_AND_LEVEL; i <= IDPB_AILERON_ROLL; i++ )
        DisableControl(i, hDLG);
    return;
}


void EnablePushbuttons(HWND hDLG)
{
    int i;

    for (i = IDPB_CONNECT; i <= IDPB_AILERON_ROLL; i++)
        EnableControl(i, hDLG);
    return;
}

void DisablePushbuttons(HWND hDLG)
{
    int i;

    for (i = IDPB_CONNECT; i <= IDPB_AILERON_ROLL; i++)
        DisableControl(i, hDLG);
    return;
}

// Disables all but the connect button.
void InitPushbuttons(HWND hDLG)
{
    DisablePushbuttons(hDLG);
    EnableControl(IDPB_CONNECT, hDLG);
    return;
}

// Sets the text displayed on/in a control.
void SetControlText(int ID, HWND hDLG, LPCSTR sz)
{
    HWND hWND;

    hWND = GetDlgItem(hDLG, ID);
    SetWindowText(hWND, sz);
    return;
}

// Fetches the text displayed on/in a control.
void GetControlText(int ID, HWND hDLG, LPTSTR sz, int Count)
{
    HWND hWND;

    hWND = GetDlgItem(hDLG, ID);
    GetWindowText(hWND, sz, Count);
    return;
}

char szTemp[20], szWorking[] = "Working...";

// Displays the string "Working..." on/in a control.  Saves the previous string.
void SetControlWorking(int ID, HWND hWND)
{
    GetControlText(ID, hWND, szTemp, 19);
    SetControlText(ID, hWND, szWorking);
    return;
}

// Restores the control's previous string.
void ResetControlString(int ID, HWND hWND)
{
   SetControlText(ID, hWND, szTemp);
   return;
}

// Converts the text value in an edit field to a double.
// Stores the number in "Number".  Returns true on success.
bool EFVal(int ID, HWND hDLG, double* Number)
{
    int Count;
    char NumStr[30];

    GetControlText(ID, hDLG, NumStr, 29);
    Count = sscanf(NumStr, "%lf", Number);
    if (Count == 1)
        return true;
    else
    {
        MessageBox(hDLG, "Please enter only numbers.", "Error", MB_OK);
        *Number = 0.0;
        return false;
    }
}


// =====================  Initialization  ===========================

// Sets the initial values in the edit fields.
void InitEditFields(HWND hDLG)
{
    SetControlText(IDEF_SERVER, hDLG, szServerName);
    SetControlText(IDEF_ALPHA, hDLG, "0.0");
    SetControlText(IDEF_SPEED, hDLG, "0.0");
    SetControlText(IDEF_DELE, hDLG, "0.0");
    SetControlText(IDEF_XPOS, hDLG, "0.0");
    SetControlText(IDEF_YPOS, hDLG, "0.0");
    SetControlText(IDEF_ALTITUDE, hDLG, "0.0");
    SetControlText(IDEF_BEARING, hDLG, "0.0");
    return;
}


// ===============  Connecting to the Simulation  ====================

// Stores the IP address in the address edit field into TCP/IP.
void SetIPADDR(HWND hDLG)
{
    HWND hWND;

    hWND = GetDlgItem(hDLG, IDEF_SERVER);
    GetWindowText(hWND, szServerName, 19);
    if (strlen(szServerName) == 0)
    {
        SetControlText(IDEF_SERVER, hDLG, "localhost");
        strcpy(szServerName, "localhost");
    }
    return;
}

void ConnectToSimulation(HWND hDLG)
{
    bool result;

    SetIPADDR(hDLG);
    SetControlText(IDPB_CONNECT, hDLG, "Working...");
    result = GetSocket();
    if (result == true)
        result &= MakeConnection();
    if (result == true)
    {
        SetControlText(IDPB_CONNECT, hDLG, "Disconnect");
        EnableTrimButtons(hDLG);
    }
    else
        SetControlText(IDPB_CONNECT, hDLG, "Connect to Server");
    return;
}

void DisconnectFromSimulation(HWND hDLG)
{
    CloseConnection();
    SetControlText(IDPB_CONNECT, hDLG, "Connect to Server");
    InitPushbuttons(hDLG);
    return;
}

void ToggleConntection(HWND hDLG)
{
    if (Connected == false)
        ConnectToSimulation(hDLG);
    else
        DisconnectFromSimulation(hDLG);
    return;
}


// ====================  Trim Data Display  =========================

// Displays the data returned in the trim response message in the edit fields.
void UpdateTrimEFields(HWND hDLG)
{
    char NumStr[20];

    if (*tr_trim_result != 0)
        SetControlText(IDSF_TRIM_STATE, hDLG, "- Trimmed -");
    else
        SetControlText(IDSF_TRIM_STATE, hDLG, "> UNTRIMMED <");
    sprintf(NumStr, "%3.3f", RadiansToDegrees(*tr_alpha));
    SetControlText(IDEF_ALPHA, hDLG, NumStr);
    sprintf(NumStr, "%3.3f", FtPerSecToKts(*tr_speed));
    SetControlText(IDEF_SPEED, hDLG, NumStr);
    sprintf(NumStr, "%3.3f", RadiansToDegrees(*tr_dele));
    SetControlText(IDEF_DELE, hDLG, NumStr);
    return;
}

// Displays the data returned in the trim response message in the static fields.
void UpdateTrimSFields(HWND hDLG)
{
    char NumStr[20];

    sprintf(NumStr, "Beta = %3.3f deg.", RadiansToDegrees(*tr_beta));
    SetControlText(IDSF_BETA_FIELD, hDLG, NumStr);
    sprintf(NumStr, "Delta A = %3.3f deg.", RadiansToDegrees(*tr_dela));
    SetControlText(IDSF_DELA_FIELD, hDLG, NumStr);
    sprintf(NumStr, "Delta R = %3.3f deg.", RadiansToDegrees(*tr_delr));
    SetControlText(IDSF_DELR_FIELD, hDLG, NumStr);
    sprintf(NumStr, "Power = %3.1f hp.", PowerToHP(*tr_power));
    SetControlText(IDSF_POWER_FIELD, hDLG, NumStr);
    sprintf(NumStr, "Throttle = %3.1f %%", *tr_throttle * 100);
    SetControlText(IDSF_THROTTLE_FIELD, hDLG, NumStr);
    return;
}

void UpdateTrimFields(HWND hDLG)
{
    UpdateTrimEFields(hDLG);
    UpdateTrimSFields(hDLG);
    return;

}

// ====================  High-Level Commands  =======================

// Fetches the values from the position, altitude and bearing fields and
// stores them in the trim command message.
bool SetTrimPosition(HWND hDLG)
{
    bool result;
    double X_Pos, Y_Pos, Alt, Bearing;

    result  = EFVal(IDEF_XPOS, hDLG, &X_Pos);
    result &= EFVal(IDEF_YPOS, hDLG, &Y_Pos);
    result &= EFVal(IDEF_ALTITUDE, hDLG, &Alt);
    result &= EFVal(IDEF_BEARING, hDLG, &Bearing);
    SetPosition(X_Pos, Y_Pos, Alt, Bearing);
    return result;
}

void SetFlyButtons(HWND hDLG)
{
    if(*tr_trim_result == 0)
        DisableFlyButtons(hDLG);
    else
        EnableFlyButtons(hDLG);
    return;
}

// Fetches the value in the trim Alpha field and sends the trim-by-alpha command.
void RunTrimAlpha(HWND hDLG)
{
    bool result;
    double Alpha;

    result = SetTrimPosition(hDLG);
    result &= EFVal(IDEF_ALPHA, hDLG, &Alpha);
    if (result == true)
    {
        TrimAlpha(Alpha);
        UpdateTrimFields(hDLG);
        SetFlyButtons(hDLG);
    }
    return;
}

// Fetches the value in the trim speed field and sends the trim-by-speed command.
void RunTrimSpeed(HWND hDLG)
{
    bool result;
    double Speed;

    result = SetTrimPosition(hDLG);
    result &= EFVal(IDEF_SPEED, hDLG, &Speed);
    if (result == true)
    {
        TrimSpeed(Speed);
        UpdateTrimFields(hDLG);
        SetFlyButtons(hDLG);
    }
    return;
}

// Fetches the value in the delE trim field and sends the trim-by-dele command.
void RunTrimDele(HWND hDLG)
{
    bool result;
    double DelE;

    result = SetTrimPosition(hDLG);
    result &= EFVal(IDEF_DELE, hDLG, &DelE);
    if (result == true)
    {
        TrimDele(DelE);
        UpdateTrimFields(hDLG);
        SetFlyButtons(hDLG);
    }
    return;
}


// ======================  Event Handlers  ==========================

// Handles commands sent by the "About" dialog box.
BOOL CALLBACK DlgAbout(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch(uMsg)
    {
    case WM_COMMAND:
            EndDialog(hwndDlg, 0);
            return true;
    }
    return false;
}


// Handles commands sent by the main dialog box.
bool ProcessCommands(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch(LOWORD(wParam))
    {
    case IDPB_CONNECT:
            ToggleConntection(hwndDlg);
            return true;
    case IDPB_TRIM_ALPHA:
            RunTrimAlpha(hwndDlg);
            return true;
    case IDPB_TRIM_SPEED:
            RunTrimSpeed(hwndDlg);
            return true;
    case IDPB_TRIM_DELE:
            RunTrimDele(hwndDlg);
            return true;
    case IDPB_STRAIGHT_AND_LEVEL:
            SetControlWorking(IDPB_STRAIGHT_AND_LEVEL, hwndDlg);
            StraightAndLevel(hwndDlg);
            ResetControlString(IDPB_STRAIGHT_AND_LEVEL, hwndDlg);
            return true;
    case IDPB_STICK_RAP:
            SetControlWorking(IDPB_STICK_RAP, hwndDlg);
            StickRap(hwndDlg);
            ResetControlString(IDPB_STICK_RAP, hwndDlg);
            return true;
    case IDPB_ELEVATOR_DOUBLET:
            SetControlWorking(IDPB_ELEVATOR_DOUBLET, hwndDlg);
            ElevatorDoublet(hwndDlg);
            ResetControlString(IDPB_ELEVATOR_DOUBLET, hwndDlg);
            return true;
    case IDPB_AILERON_DOUBLET:
            SetControlWorking(IDPB_AILERON_DOUBLET, hwndDlg);
            AileronDoublet(hwndDlg);
            ResetControlString(IDPB_AILERON_DOUBLET, hwndDlg);
            return true;
    case IDPB_RUDDER_DOUBLET:
            SetControlWorking(IDPB_RUDDER_DOUBLET, hwndDlg);
            RudderDoublet(hwndDlg);
            ResetControlString(IDPB_RUDDER_DOUBLET, hwndDlg);
            return true;
    case IDPB_LOOP:
            SetControlWorking(IDPB_LOOP, hwndDlg);
            StraightLoop(hwndDlg);
            ResetControlString(IDPB_LOOP, hwndDlg);
            return true;
    case IDPB_AILERON_ROLL:
            SetControlWorking(IDPB_AILERON_ROLL, hwndDlg);
            AileronRoll(hwndDlg);
            ResetControlString(IDPB_AILERON_ROLL, hwndDlg);
            return true;
    case MENU_ABOUT:
            DialogBox(hInst, MAKEINTRESOURCE(DLG_ABOUT), hwndDlg, (DLGPROC)DlgAbout);
            return true;
    }
    return false;
}


// The main WinProc, or in this case, DialogProc.
BOOL CALLBACK DlgMain(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    HMENU hMENU;
    bool result;
    WSADATA WsaData;

    switch(uMsg)
    {
    case WM_INITDIALOG:
            hMENU = LoadMenu(hInst, "ClientMenu");
            SetMenu(hwndDlg, hMENU);
            InitEditFields(hwndDlg);
            InitPushbuttons(hwndDlg);
            result = WSAStartup(MAKEWORD(2,2), &WsaData);
            if (result != 0)
                MessageBox(hwndDlg, "The simulation client couldn't start your computer's TCP/IP.", "Error", MB_OK);
            return TRUE;
    case WM_CLOSE:
            DisconnectFromSimulation(hwndDlg);
            WSACleanup();
            EndDialog(hwndDlg, 0);
            return TRUE;
    case WM_COMMAND:
            return ProcessCommands(hwndDlg, uMsg, wParam, lParam);
    }
    return FALSE;
}


int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
    hInst=hInstance;
    InitCommonControls();
    return DialogBox(hInst, MAKEINTRESOURCE(DLG_MAIN), NULL, (DLGPROC)DlgMain);
}

/* ====================  License Statement  =========================
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
   =============================================================== */
