\ ===================================================================
\  File: Open-Loop.FTH
\
\  This file serves as an example of how to use the command
\  words in "Commands.FTH".  It provides words to save the
\  aircraft data to an ASCII, comma-separated-value file, plus
\  open-loop control routines, of the kind that would commonly be
\  used for flight dynamics studies, that send commands to the
\  simulation to fly the simulated aircraft.
\
\              Copyright (C) 2014 Richard C. Wagner
\  This file is offered under the GNU General Public License.
\  Please see the license statement at the end of the file.
\ ===================================================================

\ ===================  ASCII File Data Output  ======================

COMMA-DELIMITED

: SAVE-TITLES ( -- )   
   S" Time, u, v, w, p, q, r, Phi, Theta, Psi, Alpha, Beta, Gamma,"
   OUTFILE-ID @ WRITE-FILE DROP
   S" X, Y, Alt., delE, delA, delR, Throttle"
   OUTFILE-ID @ WRITE-FILE DROP CR>FILE ;

: SAVE-RATE-VALUES ( -- )   ADVANCE-RESPONSE
   DUP AR-SIMTIME DF@ R>FILE
   DUP AR-u DF@ R>FILE   DUP AR-v DF@ R>FILE   DUP AR-w DF@ R>FILE 
   DUP AR-p DF@ RAD>DEG R>FILE   DUP AR-q DF@ RAD>DEG R>FILE
   AR-r DF@ RAD>DEG R>FILE ;

: SAVE-STATE-VALUES ( -- )   ADVANCE-RESPONSE
   DUP AR-Phi DF@ RAD>DEG R>FILE   DUP AR-Theta DF@ RAD>DEG R>FILE
   DUP AR-Psi DF@ RAD>DEG R>FILE
   DUP AR-ALPHA DF@ RAD>DEG R>FILE   DUP AR-BETA DF@ RAD>DEG R>FILE
   AR-GAMMA DF@ RAD>DEG R>FILE ;
   
: SAVE-POSITION-VALUES ( -- )   ADVANCE-RESPONSE
   DUP AR-x DF@ R>FILE   DUP AR-y DF@ R>FILE  AR-ALTITUDE DF@ R>FILE ;

: SAVE-DEFLECTION-VALUES ( -- )   ADVANCE-RESPONSE
   DUP AR-DELE DF@ RAD>DEG R>FILE   DUP AR-DELA     DF@ RAD>DEG R>FILE
   DUP AR-DELR DF@ RAD>DEG R>FILE       AR-THROTTLE DF@ R>FILE ;

\ Writes the current aircraft data values to the CSV file.
: SAVE-VALUES ( -- )   SAVE-RATE-VALUES  SAVE-STATE-VALUES
   SAVE-POSITION-VALUES  SAVE-DEFLECTION-VALUES  CR>FILE ;

\ Creates a CSV file named "FlightData.CSV" and writes the column labels.
\ Returns TRUE on success, FALSE if the file couldn't be created.
: CREATE-TEST ( HWND -- t )   >R
   S" FlightData.CSV" r/w CREATE-FILE SWAP OUTFILE-ID ! 0= DUP
   IF   R> DROP SAVE-TITLES
   ELSE R> Z" Couldn't create FlightData.CSV" Z" Error" MB_OK MessageBox DROP
   THEN ;

\ Closes the CSV file.
: CLOSE-TEST  ( -- )   OUTFILE-ID @ CLOSE-FILE DROP ;


\ ============  Shorthand Control Inputs and Outputs  ===============

\ Control inputs: set advance command fields
: DELA!     ( r -- )   ADVANCE-COMMAND AC-DELA DF! ;
: DELE!     ( r -- )   ADVANCE-COMMAND AC-DELE DF! ;
: DELR!     ( r -- )   ADVANCE-COMMAND AC-DELR DF! ;
: THROTTLE! ( r -- )   ADVANCE-COMMAND AC-THROTTLE DF! ;

\ Current control deflections: fetch from advance response fields
: DELA@     ( -- r )   ADVANCE-RESPONSE AR-DELA DF@ ;
: DELE@     ( -- r )   ADVANCE-RESPONSE AR-DELE DF@ ;
: DELR@     ( -- r )   ADVANCE-RESPONSE AR-DELR DF@ ;
: THRUST@   ( -- r )   ADVANCE-RESPONSE AR-THRUST DF@ ;
: THROTTLE@ ( -- r )   ADVANCE-RESPONSE AR-THROTTLE DF@ ;

\ Attitude Data: fetch from advance response fields
: PHI@ ( -- r )   ADVANCE-RESPONSE AR-PHI DF@ ;
: THETA@ ( -- r )   ADVANCE-RESPONSE AR-THETA DF@ ;
: GAMMA@ ( -- r )   ADVANCE-RESPONSE AR-GAMMA DF@ ;

\ Simulation time
: SIMTIME@ ( -- r )   ADVANCE-RESPONSE AR-SIMTIME DF@ ;


\ ==================  Open-Loop Control Routines  ===================

\ We'll use a timestep of 0.1 sec between data points.
0.1 FCONSTANT DELTA-T

\ Flies the aircraft rSEC into the future with the current control settings.
\ Saves aircraft data every DELTA-T seconds.
: FLY ( rSEC -- )   DELTA-T F/ FINT F>S
   0 DO DELTA-T ADVANCE DROP SAVE-VALUES LOOP ;

\ No control changes.  Flies straight and level for 120 seconds.
: S/L ( HWND -- )   CREATE-TEST IF 120.0 FLY CLOSE-TEST THEN ;

\ Performs a "stick rap" at 3 seconds.  Flies for a total of 120 seconds.
: STICK-RAP ( HWND -- )   CREATE-TEST
   IF 3.0 FLY
      DELE@  8.0 DEG>RAD F+ DELE!  1.0 FLY
      DELE@ -8.0 DEG>RAD F+ DELE! 116.0 FLY  CLOSE-TEST
   THEN ;

\ Performs an elevator doublet at 3 seconds.  Flies for a total of 120 seconds.
: ELEVATOR-DOUBLET ( HWND -- )   CREATE-TEST
   IF 3.0 FLY
      DELE@ -4.0 DEG>RAD F+ DELE!  1.0 FLY
      DELE@  8.0 DEG>RAD F+ DELE!  1.0 FLY
      DELE@ -4.0 DEG>RAD F+ DELE!  115.0 FLY  CLOSE-TEST
   THEN ;

\ Performs an aileron doublet at 3 seconds.  Flies for a total of 120 seconds.
: AILERON-DOUBLET ( HWND -- )   CREATE-TEST
   IF 3.0 FLY
      DELA@ -4.0 DEG>RAD F+ DELA!  1.0 FLY
      DELA@  8.0 DEG>RAD F+ DELA!  1.0 FLY
      DELA@ -4.0 DEG>RAD F+ DELA!  115.0 FLY  CLOSE-TEST
   THEN ;

\ Performs a rudder doublet at 3 seconds.  Flies for a total of 120 seconds.
: RUDDER-DOUBLET ( HWND -- )   CREATE-TEST
   IF 3.0 FLY
      DELR@ -4.0 DEG>RAD F+ DELR!  1.0 FLY
      DELR@  8.0 DEG>RAD F+ DELR!  1.0 FLY
      DELR@ -4.0 DEG>RAD F+ DELR!  115.0 FLY  CLOSE-TEST
   THEN ;


\ ===============  Semi-Open-Loop Control Routines  =================

FVARIABLE DELE-TRIM
FVARIABLE THROTTLE-TRIM

: SAVE-TRIM ( -- )   TRIM-RESPONSE
   DUP TR-DELE DF@ DELE-TRIM F!   TR-THROTTLE DF@ THROTTLE-TRIM F! ;

\ Performs a single loop at 3 seconds.  Flies until the loop is complete.
: STRAIGHT-LOOP ( HWND -- )   SAVE-TRIM  CREATE-TEST
   IF 3.0 FLY
      -15.0 DEG>RAD DELE!  1.0 THROTTLE! DELTA-T FLY
      BEGIN
         DELTA-T FLY
         GAMMA@ FDUP F0<   360.0 DEG>RAD F>= OR  SIMTIME@ 60.0 F>= OR
      UNTIL
      DELE-TRIM F@ DELE! THROTTLE-TRIM F@ THROTTLE! 10.0 FLY  CLOSE-TEST
   THEN ;

\ Performs an aileron roll at 3 seconds.  Flies until the roll is complete.
: AILERON-ROLL ( HWND -- )   CREATE-TEST
   IF 3.0 FLY
      15.0 DEG>RAD DELA! DELTA-T FLY
      BEGIN DELTA-T FLY PHI@ 360.0 DEG>RAD F>= UNTIL
      0.0 DEG>RAD DELA! 10.0 FLY  CLOSE-TEST
   THEN ;


\ ====================  License Statement  ==========================
\ This program is free software: you can redistribute it and/or modify
\ it under the terms of the GNU General Public License as published by
\ the Free Software Foundation, either version 3 of the License, or
\ (at your option) any later version.
\
\ This program is distributed in the hope that it will be useful,
\ but WITHOUT ANY WARRANTY; without even the implied warranty of
\ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
\ GNU General Public License for more details.
\
\ You should have received a copy of the GNU General Public License
\ along with this program.  If not, see <http://www.gnu.org/licenses/>.
\ ===================================================================